# -*- coding: utf-8 -*-
"""
Test utilities.
"""
from __future__ import absolute_import, unicode_literals

import os
import unittest
import tempfile

import ddt
import pgpy
from pgpy.constants import CompressionAlgorithm, HashAlgorithm, KeyFlags, PubKeyAlgorithm, SymmetricKeyAlgorithm
from pgpy.errors import PGPError

from zipfile import ZipFile
from enterprise_reporting import utils


@ddt.ddt
class TestUtilities(unittest.TestCase):
    """Utilities for Enterprise reporting scripts."""

    SCARY_DICTIONARY = {
        'key1': {
            'key1-1': [
                'key1-1-v0',
                'key1-1-v1',
                'key1-1-v2',
            ],
            'key1-2': 'key1-2-v',
            'key1-3': {
                'key1-3-1': 'key1-3-1-v',
                'key1-3-2': [
                    'key1-3-2-v0',
                    'key1-3-2-v1',
                ]
            }
        },
        'key2': [
            'key2-v0',
            'key2-v1',
        ],
        'key3': True,
        'key4': 'key4-v',
        'key5': 'key5-v',
        'key6': [
            {
                'key6-1-1': 'key6-1-1-v',
                'key6-1-2': 'key6-1-2-v',
                'key6-1-3': [
                    'key6-1-3-v0',
                    'key6-1-3-v1',
                    'key6-1-3-v2',
                ],
            },
            {
                'key6-2-1': 'key6-2-1-v',
                'key6-2-2': 'key6-2-2-v',
                'key6-2-3': [
                    'key6-2-3-v0',
                    'key6-2-3-v1',
                    'key6-2-3-v2',
                ],
            },
        ]
    }

    @ddt.data(
        (
            'key',
            SCARY_DICTIONARY,
            [
                'key1_key1-1_0',
                'key1_key1-1_1',
                'key1_key1-1_2',
                'key1_key1-2',
                'key1_key1-3_key1-3-1',
                'key1_key1-3_key1-3-2_0',
                'key1_key1-3_key1-3-2_1',
                'key2_0',
                'key2_1',
                'key3',
                'key4',
                'key5',
                'key6_0_key6-1-1',
                'key6_0_key6-1-2',
                'key6_0_key6-1-3_0',
                'key6_0_key6-1-3_1',
                'key6_0_key6-1-3_2',
                'key6_1_key6-2-1',
                'key6_1_key6-2-2',
                'key6_1_key6-2-3_0',
                'key6_1_key6-2-3_1',
                'key6_1_key6-2-3_2',
            ],
        ),
        (
            'value',
            SCARY_DICTIONARY,
            [
                'key1-1-v0',
                'key1-1-v1',
                'key1-1-v2',
                'key1-2-v',
                'key1-3-1-v',
                'key1-3-2-v0',
                'key1-3-2-v1',
                'key2-v0',
                'key2-v1',
                True,
                'key4-v',
                'key5-v',
                'key6-1-1-v',
                'key6-1-2-v',
                'key6-1-3-v0',
                'key6-1-3-v1',
                'key6-1-3-v2',
                'key6-2-1-v',
                'key6-2-2-v',
                'key6-2-3-v0',
                'key6-2-3-v1',
                'key6-2-3-v2'
            ],
        ),
    )
    @ddt.unpack
    def test_flatten_dict(self, target, dictionary, flattened):
        """The dictionary, regardless of how complex, has its keys flattened."""
        assert utils.flatten_dict(dictionary, target=target) == flattened

    @ddt.data(
        {
            'H': [
                {'H': 'i'},
                'Hi',
                {'H': 'i'},
            ],
        },
        {
            'i': [
                ['Hi']
            ],
        },
        {
            'Hi': [
                ['H'],
                ['i'],
                'Hi',
            ]
        }
    )
    def test_flatten_dict_weird_cases(self, dictionary):
        """The dictionary, when weird, raises an error."""
        with self.assertRaises(NotImplementedError):
            utils.flatten_dict(dictionary)


@ddt.ddt
class TestCompressEncrypt(unittest.TestCase):
    """
    Tests `compress_and_encrypt` works correctly.
    """
    def create_files(self, files_data):
        """
        Creates files based on provided file data.
        """
        files = []
        total_size = 0
        for file_data in files_data:
            tf = tempfile.NamedTemporaryFile(suffix='.txt')
            tf.write(file_data['size'] * b'i')
            tf.flush()
            tf.seek(0)

            files.append({
                'file': tf,
                'size': file_data['size'],
            })
            total_size += file_data['size']

        return files, total_size

    def pgpy_create_key(self, username):
        key = pgpy.PGPKey.new(PubKeyAlgorithm.RSAEncryptOrSign, 4096)
        uid = pgpy.PGPUID.new(username, comment='Unknown Person', email=username+'@unknown.com')
        key.add_uid(uid, usage={KeyFlags.Sign, KeyFlags.EncryptCommunications, KeyFlags.EncryptStorage},
                    hashes=[HashAlgorithm.SHA256, HashAlgorithm.SHA384, HashAlgorithm.SHA512, HashAlgorithm.SHA224],
                    ciphers=[SymmetricKeyAlgorithm.AES256, SymmetricKeyAlgorithm.AES192, SymmetricKeyAlgorithm.AES128],
                    compression=[CompressionAlgorithm.ZLIB, CompressionAlgorithm.BZ2, CompressionAlgorithm.ZIP,
                                 CompressionAlgorithm.Uncompressed])
        return key

    @ddt.data(
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
        ],
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
            {
                'name': 'harry-potter-and-deathly-hollowsc.txt',
                'size': 500
            },
        ],
    )
    def test_compress_and_encrypt(self, files_data):
        """
        Test that files are correctly compressed.
        """
        files, total_original_size = self.create_files(files_data)

        password = b'frodo-baggins'
        compressed_file = utils.compress_and_encrypt(
            [file['file'] for file in files],
            password
        )

        # Verify file is compressed.
        compressed_file_size = os.path.getsize(compressed_file)
        self.assertTrue(compressed_file_size < total_original_size)

        zipfile = ZipFile(compressed_file, 'r')

        for file in files:
            # Verify text file is present in zip file.
            self.assertIn(file['file'].name.split('/')[-1], zipfile.namelist())

            # Verify file content is readable is correct password.
            content = zipfile.read(file['file'].name.split('/')[-1], password)
            self.assertEqual(len(content), file['size'])

            # Also verify file is only accessible with correct password.
            with self.assertRaises(RuntimeError):
                zipfile.read(file['file'].name.split('/')[-1], b'gollum')

    @ddt.data(
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
        ],
        [
            {
                'name': 'lord-of-the-rings.txt',
                'size': 1000
            },
            {
                'name': 'harry-potter-and-deathly-hollows.txt',
                'size': 500
            },
        ],
    )
    def test_encryption(self, files_data):
        """
        Test the successful decryption with a valid key
        and an unsuccessful decryption when an incorrect key.
        """
        files, size = self.create_files(files_data)
        password = 'low-complexity'
        correct_key = self.pgpy_create_key('JohnDoe')
        wrong_key = self.pgpy_create_key('JohnDoe2')

        encrypted_file_name = utils.compress_and_encrypt(
            [file['file'] for file in files],
            password,
            str(correct_key.pubkey)
        )
        message = pgpy.PGPMessage.from_file(encrypted_file_name)
        decrypted_message = correct_key.decrypt(message)
        self.assertIsInstance(decrypted_message, pgpy.PGPMessage)

        with self.assertRaises(PGPError):
            wrong_key.decrypt(message)



"""
Permissions needed to restrict access to the enterprise data api.
"""
from __future__ import absolute_import, unicode_literals

from logging import getLogger

from rest_framework import permissions

from enterprise_data.clients import EnterpriseApiClient

LOGGER = getLogger(__name__)


class IsStaffOrEnterpriseUser(permissions.BasePermission):
    """
    Permission that checks to see if the request user is staff or is associated with the enterprise in the request.

    NOTE: This permission check may make a request to the LMS to get the enterprise association if it is not already in
    the session. This fetch should go away when JWT Scopes are fully implemented and the association is stored on
    the token.
    """

    ENTERPRISE_DATA_API_GROUP = 'enterprise_data_api_access'

    def get_user_enterprise_data(self, auth_token, user):
        """
        Get the enterprise learner model from the LMS for the given user.

        Returns: learner or None if unable to get or user is not associated with an enterprise
        """
        enterprise_client = EnterpriseApiClient(auth_token)
        enterprise_learner_data = enterprise_client.get_enterprise_learner(user)
        if not enterprise_learner_data:
            return None

        return {
            'enterprise_id': enterprise_learner_data['enterprise_customer']['uuid'],
            'enterprise_groups': enterprise_learner_data['groups'],
        }

    def has_permission(self, request, view):
        """
        Verify the user is staff or the associated enterprise matches the requested enterprise.
        """
        if request.user.is_staff:
            return True

        if not hasattr(request.session, 'enterprise_id') or not hasattr(request.session, 'enterprise_groups'):
            enterprise_data = self.get_user_enterprise_data(request.auth, request.user)
            if not enterprise_data:
                return False
            request.session['enterprise_id'] = enterprise_data['enterprise_id']
            request.session['enterprise_groups'] = enterprise_data['enterprise_groups']

        enterprise_in_url = request.parser_context.get('kwargs', {}).get('enterprise_id', '')

        permitted = (
            request.session['enterprise_id'] == enterprise_in_url and
            self.ENTERPRISE_DATA_API_GROUP in request.session['enterprise_groups']
        )
        if not permitted:
            LOGGER.warning('User {} denied access to EnterpriseEnrollments for enterprise {}'
                           .format(request.user, enterprise_in_url))

        return permitted


class HasDataAPIDjangoGroupAccess(permissions.BasePermission):
    """
    Permission that checks to see if the request user is part of the enterprise_data_api django group.

    Also checks that the user is authorized for the request's enterprise.
    """

    def get_enterprise_with_access_to(self, auth_token, user, enterprise_id):
        """
        Get the enterprise customer data that the user has enterprise_data_api access to.

        Returns: enterprise or None if unable to get or user is not associated with an enterprise
        """
        enterprise_client = EnterpriseApiClient(auth_token)
        enterprise_data = enterprise_client.get_with_access_to(user, enterprise_id)
        if not enterprise_data:
            return None

        return enterprise_data

    def has_permission(self, request, view):
        """
        Verify the user is staff or the associated enterprise matches the requested enterprise.
        """
        enterprise_in_url = request.parser_context.get('kwargs', {}).get('enterprise_id', '')

        if ('enterprises_with_access' not in request.session or
                enterprise_in_url not in request.session['enterprises_with_access']):
            request.session['enterprises_with_access'] = {}
            enterprise_data = self.get_enterprise_with_access_to(request.auth, request.user, enterprise_in_url)
            if not enterprise_data:
                request.session['enterprises_with_access'][enterprise_in_url] = False
            else:
                request.session['enterprises_with_access'][enterprise_in_url] = True

        permitted = request.session['enterprises_with_access'][enterprise_in_url]
        if not permitted:
            LOGGER.warning('User {} denied access to EnterpriseEnrollments for enterprise {}'
                           .format(request.user, enterprise_in_url))

        return permitted



# -*- coding: utf-8 -*-
"""
Serializers for enterprise api version 0 endpoint.
"""
from __future__ import absolute_import, unicode_literals

from rest_framework import serializers

from enterprise_data.models import EnterpriseEnrollment, EnterpriseUser


class EnterpriseEnrollmentSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseEnrollment model.
    """
    course_api_url = serializers.SerializerMethodField()

    def get_course_api_url(self, obj):
        """Constructs course api url"""
        return '/enterprise/v1/enterprise-catalogs/{enterprise_id}/courses/{course_id}'.format(
            enterprise_id=obj.enterprise_id, course_id=obj.course_id
        )

    class Meta:
        model = EnterpriseEnrollment
        exclude = ('created', )


class EnterpriseUserSerializer(serializers.ModelSerializer):
    """
    Serializer for EnterpriseUser model.
    """

    class Meta:
        model = EnterpriseUser
        exclude = ('created', )

    def to_representation(self, instance):
        representation = super(EnterpriseUserSerializer, self).to_representation(instance)

        if hasattr(instance, 'enrollment_count'):
            representation['enrollment_count'] = instance.enrollment_count
        if hasattr(instance, 'course_completion_count'):
            representation['course_completion_count'] = instance.course_completion_count

        return representation


class LearnerCompletedCoursesSerializer(serializers.Serializer):    # pylint: disable=abstract-method
    """
    Serializer for learner's completed courses.
    """
    user_email = serializers.EmailField()
    completed_courses = serializers.IntegerField()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2018-06-28 16:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0008_auto_20180614_0108'),
    ]

    operations = [
        migrations.RenameField(
            model_name='enterpriseenrollment',
            old_name='final_grade',
            new_name='current_grade',
        ),
        migrations.AddField(
            model_name='enterpriseenrollment',
            name='offer',
            field=models.CharField(max_length=128, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2018-03-02 08:47
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='EnterpriseEnrollment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('enterprise_id', models.CharField(max_length=32)),
                ('enterprise_name', models.CharField(max_length=255)),
                ('lms_user_id', models.PositiveIntegerField()),
                ('enterprise_user_id', models.PositiveIntegerField()),
                ('course_id', models.CharField(help_text='The course the learner is enrolled in.', max_length=255)),
                ('enrollment_created_timestamp', models.DateTimeField()),
                ('user_current_enrollment_mode', models.CharField(max_length=32)),
                ('consent_granted', models.BooleanField(default=False)),
                ('letter_grade', models.CharField(max_length=32, null=True)),
                ('has_passed', models.BooleanField(default=False)),
                ('passed_timestamp', models.DateTimeField(null=True)),
                ('enterprise_sso_user_id', models.CharField(max_length=255, null=True)),
                ('course_title', models.CharField(max_length=255, null=True)),
                ('course_start', models.DateTimeField(null=True)),
                ('course_end', models.DateTimeField(null=True)),
                ('course_pacing_type', models.CharField(max_length=32, null=True)),
                ('course_duration_weeks', models.CharField(max_length=32, null=True)),
                ('course_min_effort', models.PositiveIntegerField(null=True)),
                ('course_max_effort', models.PositiveIntegerField(null=True)),
                ('user_account_creation_date', models.DateTimeField(null=True)),
                ('user_email', models.CharField(max_length=255, null=True)),
                ('user_username', models.CharField(max_length=255, null=True)),
            ],
            options={
                'db_table': 'enterprise_enrollment',
                'verbose_name': 'Enterprise Enrollment',
                'verbose_name_plural': 'Enterprise Enrollments',
            },
        )
    ]



"""
Filters for enterprise data views.
"""
from __future__ import absolute_import, unicode_literals

from rest_framework import filters

from django.db.models import Q

# Admittedly this is sort of hacky because the use of "|" with 2 Q objects
# forces the ORM to use a LEFT OUTER JOIN, which is needed to return a user
# that has multiple enrollments where at least one has consent_granted=True
CONSENT_TRUE_OR_NOENROLL_Q = Q(enrollments__consent_granted=True) | Q(enrollments__isnull=True)


class ConsentGrantedFilterBackend(filters.BaseFilterBackend):
    """
    Filter backend for any view that needs to filter results where consent has not been granted.

    This requires that `CONSENT_GRANTED_FILTER` be set in the view as a class variable, to identify
    the object's relationship to the consent_granted field.
    """

    def filter_queryset(self, request, queryset, view):
        """
        Filter a queryset for results where consent has been granted.
        """
        filter_kwargs = {view.CONSENT_GRANTED_FILTER: True}
        return queryset.filter(**filter_kwargs)
